//
// juma-uart2.c
//
//
// quotation from Simple UART interface for dsPIC30F6012
// Juha Niinikoski, OH2NLT 18.05.2005
//
//
//	ver.0.20 15 Mar 2015
//	rx buffer fill flag added (for accurate string aquisition from GPS)
//
//	ver.0.10 9 Mar 2015
//	first implementation
//


#include "juma-tx500.h"
#include <stdio.h>

// Baud Rate divisors for 29,4912 MHz osc =  7,3728 MHz FCY 
/* actual baud rate divisors
[0] 1200	383
[1] 2400	191
[2] 4800	95
[3] 9600	47
[4] 19200	23
[5] 38400	11
[6] 57600	7
[7] 115200	3
*/

volatile unsigned char rx_buf[64];
volatile unsigned char in_point = 0;
volatile unsigned char buf_fill = 0;		// when reach the end of rx buffer index
volatile unsigned char gl[7];				// calculated grid locator(6digits)
volatile unsigned char gps_active = 0;		// when uart2 interrupt occur

// 7,3728MHz
const unsigned int baud_rates2[8] = {383, 191, 95, 47, 23, 11, 7, 3};

// UART2 IRQ service
// for GPS NMEA format
void __attribute__((interrupt, auto_psv)) _U2RXInterrupt(void)  // put receiced character to RX queue
	{
  	unsigned char rx;

  	rx = U2RXREG;                        			// data read
	if(rx == '$') in_point = 0;						// reset index when GPS sentence starting character get

	if(in_point < 64){
		rx_buf[in_point] = rx;
		in_point++;
		buf_fill = 0;
	}
	else buf_fill = 1;								// data buffer filled
	
	if(U2STAbits.OERR == 1)							// OERR is blocking uart, should not happen but..
		U2STAbits.OERR = 0;							// reset overrun error if occured

// this is not automatic in 30F6014 UART
	IFS1bits.U2RXIF = 0;							// reset irq flag
	gps_active = 1;
	}


// UART2 init
void InitUSART2(void)
{
 U2BRG = baud_rates2[2];				// set intial baud rate to 9600 (temporal)

 U2MODE = 0x0000;
 U2STA = 0x0000;						// IRQ on every character
 IFS1bits.U2RXIF = 0;					// clear possible pending irq
 IEC1bits.U2RXIE = 1;					// enable RX IRQs
 U2MODEbits.UARTEN = 1;					// enable uart
 U2STAbits.UTXEN = 1;					// enable transmitter
}

// set baud rate
void SetUSART2baud(int baud_sel)
	{
	if(baud_sel > 7)
		baud_sel = 7;					// stay in the table

	 U2BRG = baud_rates2[baud_sel];		// set baud rate
	}

// print character
void putch2(char c)
	{
	while(U2STAbits.UTXBF == 1)			// wait here if tx fifo full
		{
		}
	U2TXREG = (int)c &0x00FF;			// Tx reg is 16-bit long
	}

void putchars(const unsigned char *str){
    unsigned char i = 0;
    while(str[i] != 0){
        putch2(str[i]);
        i++;
    }
}

void puthex(unsigned char i){
    unsigned char j;
    j = i >> 4;
    if(j < 10) putch2(j + 0x30);
    else putch2(j + 0x37);
    j = i & 0x0F;
    if(j < 10) putch2(j + 0x30);
    else putch2(j + 0x37);
}

void calc_gl(void){
	unsigned char temp;
	unsigned int temp2;

	temp = (rx_buf[30] - '0') * 10 + (rx_buf[31] - '0');
	temp2 = (((rx_buf[32]-'0') % 2)*6+(rx_buf[33]-'0'))*1000+(rx_buf[34]-'0')*100+(rx_buf[36]-'0')*10+(rx_buf[37]-'0');
	if(rx_buf[41] == 'E'){							// EAST side
		temp = temp + 18;
		gl[0] = (temp >> 1) + 'A';
		temp = (temp % 2) * 10 + (rx_buf[32] - '0');
		gl[2] = (temp >> 1) + '0';
	}
	else{											// WEST side
		temp = 18 - temp;
		if((rx_buf[32] + rx_buf[33] + rx_buf[34] + rx_buf[36] + rx_buf[37]) & 0x0F)
			gl[0] = ((temp - 1) >> 1) + 'A';
		else
			gl[0] = (temp >> 1) + 'A';
		temp = (temp % 2) * 10 - (rx_buf[32] - '0');
		if((rx_buf[33] + rx_buf[34] + rx_buf[36] + rx_buf[37]) & 0x0F)
			gl[2] = ((temp - 1) >> 1) + '0';
		else
			gl[2] = (temp >> 1) + '0';
		temp2 = 12000 - temp2;
	}
	gl[4] = (unsigned char)(temp2 / 500) + 'a';

	temp = rx_buf[18] - '0';
	temp2 = (rx_buf[20]-'0')*1000+(rx_buf[21]-'0')*100+(rx_buf[23]-'0')*10+(rx_buf[24]-'0');		
	if(rx_buf[28] == 'N'){							// NORTH side
		temp = temp + 9;
		gl[1] = temp + 'A';
		gl[3] = rx_buf[19];
	}
	else{											// SOUTH side
		temp = 9 - temp;
		if((rx_buf[19] + rx_buf[20] + rx_buf[21] + rx_buf[23] + rx_buf[24]) & 0x0F)
			gl[1] = temp - 1 + 'A';
		else
			gl[1] = temp + 'A';
		temp = 10 - (rx_buf[19] - '0');
		if((rx_buf[20] + rx_buf[21] + rx_buf[23] + rx_buf[24]) & 0x0F)
			gl[3] = temp - 1 + '0';
		else
			gl[3] = temp + '0';
		temp2 = 6000 - temp2;
	}
	gl[5] = (unsigned char)(temp2 / 250) + 'a';
	gl[6] = 0;
}
