//
//	juma-wspr.c
// 
//	WSPR playing, string edit and symbol encoding program for JUMA TX-136/500 by JL1VNQ / HARU
//
//
//	ver.0.70 24 Mar 015
//	start point 00:00 --> 00:01
//
//	ver.0.66 20 Mar 2015
//	enc_wspr() routine bug fixed (wspr_char[i] &= 0x0F --> wspr_char[i] & 0x0F)
//
//	ver.0.65 19 Mar 2015
//	one symbol sending delay re-adjustment (1ms faster on WSPR2 mode, 6ms faster on WSPR15 mode)
//
//	ver.0.50 27 Feb 2015
//	bugs fix(detect EEPROM initial data 0xFF, re-display when push Power off switch breaking)
//	code arrangement, string select limit(ex. dBm 0 to 60, 037 step)
//
//	ver.0.40 18 Feb 2015
//	last saved string edit (no all clear)
//
//	ver.0.30 23 Jan 2015
//	wspr_busy_counter (= 1 / baud rate) tuning
//
//	ver.0.20 21 Jan 2015
//	abnormality when display with "wspr_char" array fixed (new variable array "wspr_disp_char")
//	implemented center frequency offset
//	minor bug fix
//
//	ver.0.10 15 Jan 2015
//	first implementation(WSPR2 mode manual playing, string edit, symbol encoding)
//

#include "juma-tx500.h"
#include "DataEEPROM.h"
#include "lcd-tx500.h"
#include "tx500_eeprom.h"

volatile unsigned char wspr_char[13];
extern const char char_select[56];							// from juma-qrss.c

volatile char wspr_mem_buf[13] = {0};

volatile char lsym[41] = {0};

const unsigned char sync[41] = {							// virtual sync symbol(format "0b0x0x0x0x")
    0b01010000,0b00000000,0b01000000,0b01010100,0b00000100,0b00010001,0b01010100,0b00000000,
    0b00000100,0b00010001,0b00000000,0b00000100,0b01010000,0b01010001,0b00000001,0b01000100,
    0b00000001,0b01000100,0b01000100,0b01000001,0b00000100,0b01010000,0b00010100,0b01000100,
    0b00000100,0b00000000,0b01000001,0b00000101,0b01000101,0b00000101,0b00010000,0b00010101,
    0b00000000,0b00010001,0b00000101,0b00000000,0b00000001,0b01000100,0b01010000,0b00010100,
    0b00000000
};

extern unsigned int wspr_playflag;							// 1:in play
extern unsigned int wspr_period[4];							// WSPR period table
extern int p_idx;											// period table index

unsigned int wspr_editflag = 0;								// 1:in edit
unsigned long wspr_offset = 0;								// center frequency offset value
unsigned int wspr_busy_counter = 0;							// msec(max 32.768sec)

char lcd_wspr_buf[10] = {"         "};

int enc_wspr(void);

//extern int wp;											// wspr encoder success flag

extern unsigned long tw_shift;								// tword shift level
extern int tx_mode;

extern int wspr_tim;

extern int str_edit;
extern int lcd_mode;
extern int edit_break;

extern int board_type;
extern int sub_page1;

extern void tx_on(void);
extern void tx_off(void);

extern void cur_on(void);
extern void cur_off(void);

extern unsigned int long_push;
extern void ms_delay(unsigned int);

extern void	clk_lcd(unsigned char);
extern void wait_lcd_rdy(void);
extern void set_cur_lcd(unsigned char);
extern void clear_lcd(void);
extern void lcd_putch(unsigned char);
extern void lcd_putst(register const char *);

extern int encoder_get(void);

extern void beep(int, unsigned int);

extern union
	{
	struct defval defval;
	unsigned int storage[sizeof(struct defval)/2];
	}eeprom;

extern union
	{
	struct calval calval;
	unsigned int ee[sizeof(struct calval)/2];
	}cal;

// in 1msec irq

void wspr_1ms_counter(void){
	if(wspr_busy_counter != 0) wspr_busy_counter--;
}

// in main routine

void save_wspr_string(void){
	unsigned int x, y, temp;
	for(x=0;x<12;x++){
		temp = (unsigned int)wspr_mem_buf[x];
		y = EraseEE(EEPAGE, (2*x+EEWSPR), WORD);			// EEPAGE = 0x7F, WORD = 1 (DataEEPROM.h)
		y = WriteEE(&temp, EEPAGE, (2*x+EEWSPR), WORD);		// EEPROM Address 8 high bits, address + physical EEPROM start 16 low bits
	}
}

void read_wspr_string(void){								// startup only
	unsigned int x,y,temp;

	for(x=0;x<12;x++){
		y = ReadEE(EEPAGE, (2*x+EEWSPR), &temp, WORD);
		wspr_char[x] = (char)temp;
		}
}

void erase_wspr_string(void){
	unsigned int x, y;
	for(x=0;x<12;x++){
		y = EraseEE(EEPAGE, (2*x+EEWSPR), WORD);			// EEPAGE = 0x7F, WORD = 1 (DataEEPROM.h)
	}
}

void wspr_lcd_curset(char idx){
	wait_lcd_rdy( );										// check lcd busy flag
	LCD_RS = 0;												// switch command mode
	if(idx < 6) clk_lcd(0x89 + idx);
	else if(idx < 10) clk_lcd(0xBD + idx);
	else if(idx < 12) clk_lcd(0xC3 + idx);
	LCD_RS = 1;												// back to data mode
}
//
// WSPR playing routine
//

void wspr_play(void){
	static unsigned int sym_cnt = 0;						// symbol counter(total 162symbols)
	static unsigned char _1sym_outflag = 0;					// one symbol playing flag
	char p_sym;												// in play symbol value
	unsigned int i;
	unsigned long shift_unit;

		if(board_type == 1){
			if(wspr_period[p_idx] == 2) shift_unit = 131;		// (12000/8192) / (6,000,000/2/2^28) = 131.072(WSPR2)
			if(wspr_period[p_idx] == 15) shift_unit = 16;		// /8 = 16.384(WSPR15)
		}
		if(board_type == 2){
			if(wspr_period[p_idx] == 2) shift_unit = 39;		// (12000/8192) / (20,000,000/2/2^28) = 39.3216(WSPR2)
			if(wspr_period[p_idx] == 15) shift_unit = 5;		// /8 = 4.9152(WSPR15)
		}
	
		wspr_offset = (shift_unit * 3) >> 1;					// center frequency offset calculation

	if(wspr_playflag != 0){
		if((wspr_busy_counter == 0) && (_1sym_outflag == 0)){
			p_sym = 0x03 & ((lsym[sym_cnt/4]) >> ((6 - (sym_cnt % 4) * 2)));
			tw_shift = shift_unit * (3 - p_sym);
			if(wspr_period[p_idx] == 2)
//				wspr_busy_counter = 683;					// 1.46484375 baud(WSPR2)
				wspr_busy_counter = 679;					// 4msec offset
			if(wspr_period[p_idx] == 15)
//				wspr_busy_counter = 5461;					// 0.1831 baud(WSPR15)
				wspr_busy_counter = 5441;					// 20msec offset
			_1sym_outflag = 1;
			tx_on();
			sprintf(lcd_wspr_buf," [%d]%3d  ", p_sym, (162 - sym_cnt));
		}
		else if((_1sym_outflag == 1) && (wspr_busy_counter == 0)){
			if(sym_cnt < 162){
				_1sym_outflag = 0;
				sym_cnt++;
			}
			else{
				_1sym_outflag = 0;
				tw_shift = 0;
				wspr_offset = 0;
				tx_off();
				sym_cnt = 0;
				wspr_playflag = 0;
			}
		}
	}
	else{
		_1sym_outflag = 0;
		tw_shift = 0;
		wspr_offset = 0;
		tx_off();
		sym_cnt = 0;
		for(i=0;i<8;i++){
			lcd_wspr_buf[i] = ' ';
		}
	}
}

//
// WSPR string character edit routine
//

void wspr_edit(void){
	
	static char buf_idx = 0;
	static char cs_idx = 0;
	int i,temp;

	if(wspr_editflag == 0){

		wspr_editflag = 1;									// only one time

		clear_lcd();
		
		set_cur_lcd(LINE1);
		lcd_putst("Callsign[      ]");
		set_cur_lcd(LINE2);
		lcd_putst("GL[    ] dBm[  ]");
		
		for(i=0;i<12;i++){									// string copy from playing buffer
			wspr_mem_buf[i] = wspr_char[i];
		}
		
		if((wspr_mem_buf[0] + 1) > 1){						// wspr_mem_buf[0] except 0xFF or 0x00

			for(i=0;i<12;i++){								// display saved string
				wspr_lcd_curset(i);
				lcd_putch(wspr_char[i]);
			}

			wspr_lcd_curset(11);
			cur_on();
			
			for(i=0;i<56;i++){
				if(char_select[i] == wspr_mem_buf[11]) cs_idx = i;
			}
			buf_idx = 11;
		}
		else{
			wspr_mem_buf[0] = char_select[0];				// initial display "A"

			wspr_lcd_curset(0);
			lcd_putch('A');

			wspr_lcd_curset(0);
			cur_on();

			buf_idx = 0;
		}
	}
	
	if(edit_break){
		clear_lcd();
		
		set_cur_lcd(LINE1);
		lcd_putst("Callsign[      ]");
		set_cur_lcd(LINE2);
		lcd_putst("GL[    ] dBm[  ]");
		
		for(i=0;i<buf_idx+1;i++){
			wspr_lcd_curset(i);
			lcd_putch(wspr_mem_buf[i]);
		}
		
		wspr_lcd_curset(buf_idx);
		cur_on();
		
		edit_break = 0;
	}

	temp = encoder_get();

	if(temp != 0){
		
		cur_off();
		
		cs_idx = cs_idx + temp;
		
		switch(buf_idx){
			case 0:
				if(cs_idx < 0) cs_idx = 37;
				if(cs_idx > 37) cs_idx = 0;
				break;
			case 1: case 3: case 4:
				if(cs_idx < 0) cs_idx = 25;
				if(cs_idx > 25) cs_idx = 0;
				break;
			case 5:
				if(cs_idx < 0) cs_idx = 26;
				if(cs_idx > 26) cs_idx = 0;
				break;
			case 6: case 7:
				if(cs_idx < 0) cs_idx = 17;
				if(cs_idx > 17) cs_idx = 0;
				break;
			case 2: case 8: case 9:
				if(cs_idx < 27) cs_idx = 36;
				if(cs_idx > 36) cs_idx = 27;
				break;
			case 10:
				if(cs_idx < 27) cs_idx = 33;
				if(cs_idx > 33) cs_idx = 27;
				break;
			case 11:
				if(cs_idx < 27 || cs_idx == 31) cs_idx = 34;
				if(cs_idx > 34 || cs_idx == 29) cs_idx = 27;
				if(cs_idx == 28 || cs_idx == 33) cs_idx = 30;
				if(wspr_mem_buf[10] == '6') cs_idx = 27;
				break;
		}

//		if(cs_idx < 0) cs_idx = 55;
//		if(cs_idx > 55) cs_idx = 0;
		
		wspr_mem_buf[(unsigned char)buf_idx] = char_select[(unsigned char)cs_idx];
		
		wspr_lcd_curset(buf_idx);
		lcd_putch(wspr_mem_buf[(unsigned char)buf_idx]);
		
		wspr_lcd_curset(buf_idx);							// cursor location reset
		cur_on();
	}

	if(DISP == 0)
		{
		long_push = LONG_PUSH;
		
		while(DISP == 0){
			if(long_push == 0){
				beep(TONE_C, cal.calval.beep_len*10);		// long beep for save
				break;
			}
		}
		
		if(long_push == 0){									// for save message
			if(buf_idx == 11){
				save_wspr_string();
				for(i=0;i<13;i++){
					wspr_char[i] = wspr_mem_buf[i];			// string copy to playing & display buffer
				}
				
				clear_lcd();

				cur_off();

				set_cur_lcd(LINE1);
				lcd_putst("New WSPR String ");
				set_cur_lcd(LINE2);
				lcd_putst("Saved to EEPROM!");
				ms_delay(2000);
				
				enc_wspr();									// WSPR symbols encode
				
				buf_idx = 0;
				cs_idx = 0;
				str_edit = 0;
				lcd_mode = 0;
				sub_page1 = 0;
				wspr_editflag = 0;
				
				for(i=0;i<12;i++){							// clear string buffer
					wspr_mem_buf[i] = 0;
				}
			}
			else{
				clear_lcd();
				
				cur_off();
				
				set_cur_lcd(LINE1);
				lcd_putst("  Not complete  ");
				set_cur_lcd(LINE2);
				lcd_putst(" Return to edit ");
				ms_delay(2000);
				
				edit_break = 1;
			}
		}													// to exit
		else{												// for definition character
			if(buf_idx < 11){

				cur_off();

				buf_idx++;
				switch(buf_idx){
					case 2: case 8: case 10: case 11:
						cs_idx = 27;
						break;
					case 3:
						cs_idx = 0;
						break;
					}
				
				wspr_mem_buf[(unsigned char)buf_idx] = char_select[(unsigned char)cs_idx];
				
				wspr_lcd_curset(buf_idx);
				lcd_putch(wspr_mem_buf[(unsigned char)buf_idx]);
				
				wspr_lcd_curset(buf_idx);					// cursor location reset
				cur_on();
			}
		}
	}

	if(OPER == 0)
		{
		long_push = LONG_PUSH;
		
		while(OPER == 0){
			if(long_push == 0){
				beep(TONE_C, cal.calval.beep_len*10);		// long beep for cancel
				break;
			}
		}
		
		if(long_push == 0){									// for save cancel message

			clear_lcd();

			cur_off();

			set_cur_lcd(LINE1);
			lcd_putst(" Save Cancelled ");
			set_cur_lcd(LINE2);
			lcd_putst("String no change");
			ms_delay(2000);
			
			buf_idx = 0;
			cs_idx = 0;
			str_edit = 0;
			lcd_mode = 0;
			sub_page1 = 0;
			wspr_editflag = 0;
			
			for(i=0;i<12;i++){								// clear string buffer
				wspr_mem_buf[i] = 0;
			}
		}													// to exit
		else{												// for back space
			if(buf_idx > 0){

				cur_off();

				wspr_mem_buf[(unsigned char)buf_idx] = 0;

				wspr_lcd_curset(buf_idx);
				lcd_putch(' ');
				
				buf_idx--;

				wspr_lcd_curset(buf_idx);					// cursor location reset
				cur_on();

				for(i=0;i<56;i++){							// char index search
					if(char_select[i] == wspr_mem_buf[(unsigned char)buf_idx]) cs_idx = i;
				}
			}
		}
	}
}

//
// WSPR symbol encoding routine from picwsprgen.c
//


int enc_wspr(void){
    unsigned char num_char[12] = {0};
    unsigned long comp[2] = {0};

    unsigned long shift_reg = 0;
    unsigned char sym0 = 0;
    unsigned char sym1 = 0;

	unsigned char i_rev = 0;
	unsigned char p = 0;

	unsigned char i,j;

/////////////// clear buffer ///////////////

    for(i=0; i<41; i++) {
        lsym[i] = 0;
    }

/////////////// convert characters to numerical values ///////////////

    for(i=0; i<12; i++) {
        if((wspr_char[i] >= 0x30) && (wspr_char[i] <= 0x39)) num_char[i] = (wspr_char[i] & 0x0F);
        else if((wspr_char[i] >= 0x41) && (wspr_char[i] <= 0x5A)) num_char[i] = wspr_char[i] - 'A' + 10;
        else if((wspr_char[i] >= 0x61) && (wspr_char[i] <= 0x7A)) num_char[i] = wspr_char[i] - 'a' + 10;
        else if(wspr_char[i] == 0x20) num_char[i] = 36;
        else return 1;				// detect error of illegal input characters

        if((i >= 6) && (num_char[i] > 9)) num_char[i] -= 10;
    }

    if((num_char[1] == 36) || (num_char[2] > 9) || (num_char[3] < 10) || (num_char[4] < 10) || (num_char[5] < 10))
    	return 1;					// detect error of illegal input characters

	comp[0] = num_char[0]*36;
	comp[0] = (comp[0] + num_char[1])*10;
	comp[0] = (comp[0] + num_char[2])*27;
	comp[0] = (comp[0] + (num_char[3] - 10))*27;
	comp[0] = (comp[0] + (num_char[4] - 10))*27;
	comp[0] = comp[0] + (num_char[5] - 10);
	comp[1] = (179 - (num_char[6]*10 + num_char[8]))*180;
	comp[1] = (comp[1] + num_char[7]*10 + num_char[9])*128;
	comp[1] = comp[1] + num_char[10]*10 + num_char[11] + 64;

    comp[0] = (comp[0] << 4) | (comp[1] >> 18);
    comp[1] <<= 14;

/////////////// convolver encoding (k=32, r=1/2) ///////////////

   for(i=0; i<81; i++) {
        shift_reg <<= 1;

        if(i < 32){
            if((comp[0] << i) & 0x80000000) shift_reg |= 1;
        }
        else{
            if((comp[1] << (i - 32)) & 0x80000000) shift_reg |= 1;
        }

        for(j=0; j<32; j++) {
            sym0 ^= (((shift_reg & 0xf2d05351) >> j) & 0x00000001);
            sym1 ^= (((shift_reg & 0xe4613c47) >> j) & 0x00000001);
        }

        if((i % 2) == 0) lsym[i >> 1] |= (((sym0 & 0x01) << 3) | ((sym1 & 0x01) << 2));
        else lsym[i >> 1] |= (((sym0 & 0x01) << 1) | (sym1 & 0x01));

        sym0 = 0;
        sym1 = 0;
    }

/////////////// interleave & merge with sync symbols ///////////////

	if(p < 162) {
		for(i=0; i<255; i++) {
			for(j=0; j<8; j++) {
				i_rev |= (((i >> j) & 0x01) << (7 - j));
			}
			if(i_rev < 162) {
				lsym[i_rev >> 2] |= (((lsym[p >> 2] >> (3 - (p % 4))) & 0x01) << (7 - (i_rev % 4)));
				p++;
			}
			i_rev = 0;
		}
	}

    for(i=0; i<41; i++){
		lsym[i] &= 0xF0;
        lsym[i] = ((lsym[i] & 0x10) >> 3) | ((lsym[i] & 0x20) >> 2) | ((lsym[i] & 0x40) >> 1) | (lsym[i] & 0x80);
		lsym[i] |= sync[i];				// merge with sync symbol
    }
    return 0;
}
